@@ -27,7 +27,8 @@ enum {
 	NET_BH,
 	IMMEDIATE_BH,
 	KEYBOARD_BH,
-	CYCLADES_BH
+	CYCLADES_BH,
+	CM206_BH
 };
 
 extern inline void mark_bh(int nr)
@@ -77,14 +78,14 @@ extern inline void end_bh_atomic(void)
  * 7. service the device to clear its pending interrupt.
  * 8. loop again if paranoia is required.
  *
- * probe_irq_on() returns a mask of snarfed irq's.
+ * probe_irq_on() returns a mask of allocated irq's.
  *
  * probe_irq_off() takes the mask as a parameter,
  * and returns the irq number which occurred,
  * or zero if none occurred, or a negative irq number
  * if more than one irq occurred.
  */
-extern unsigned int probe_irq_on(void);	/* returns 0 on failure */
-extern int probe_irq_off(unsigned int); /* returns 0 or negative on failure */
+extern unsigned long probe_irq_on(void);	/* returns 0 on failure */
+extern int probe_irq_off(unsigned long);	/* returns 0 or negative on failure */
 
 #endif
