@@ -71,12 +71,11 @@ struct vm_area_struct {
 
 #define VM_GROWSDOWN	0x0100	/* general info on the segment */
 #define VM_GROWSUP	0x0200
-#define VM_SHM		0x0400
+#define VM_SHM		0x0400	/* shared memory area, don't swap out */
 #define VM_DENYWRITE	0x0800	/* ETXTBSY on write attempts.. */
 
 #define VM_EXECUTABLE	0x1000
-#define VM_DONTSWAP	0x2000  /* Some vm types have their own
-				 * hard-coded swap mechanism */
+#define VM_LOCKED	0x2000
 
 #define VM_STACK_FLAGS	0x0177
 
@@ -107,27 +106,20 @@ struct vm_operations_struct {
 	pte_t (*swapin)(struct vm_area_struct *, unsigned long, unsigned long);
 };
 
-extern mem_map_t * mem_map;
-extern unsigned char *age_map;
-
-/* planning stage.. */
-#define P_DIRTY		0x0001
-#define P_LOCKED	0x0002
-#define P_UPTODATE	0x0004
-#define P_RESERVED	0x8000
-
-struct page_info {
-	unsigned short flags;
-	unsigned short count;
-	struct inode * inode;
+typedef struct page {
+	unsigned int count;
+	unsigned dirty:16,
+		 age:8,
+		 unused:7,
+		 reserved:1;
 	unsigned long offset;
-	struct page_info * next_same_inode;
-	struct page_info * prev_same_inode;
-	struct page_info * next_hash;
-	struct page_info * prev_hash;
-	struct wait_queue *wait;
-};
-/* end of planning stage */
+	struct inode *inode;
+	struct page *write_list;
+	struct page *next, *prev;
+	struct page *next_hash, *prev_hash;
+} mem_map_t;
+
+extern mem_map_t * mem_map;
 
 /*
  * Free area management
@@ -145,7 +137,7 @@ struct mem_list {
 };
 
 extern struct mem_list free_area_list[NR_MEM_LISTS];
-extern unsigned char * free_area_map[NR_MEM_LISTS];
+extern unsigned int * free_area_map[NR_MEM_LISTS];
 
 /*
  * This is timing-critical - most of the time in getting a new page
@@ -179,13 +171,13 @@ extern void free_page_tables(struct task_struct * tsk);
 extern void clear_page_tables(struct task_struct * tsk);
 extern int new_page_tables(struct task_struct * tsk);
 extern int copy_page_tables(struct task_struct * to);
-extern int clone_page_tables(struct task_struct * to);
-extern int unmap_page_range(unsigned long from, unsigned long size);
+
+extern int zap_page_range(struct mm_struct *mm, unsigned long address, unsigned long size);
+extern int copy_page_range(struct mm_struct *dst, struct mm_struct *src, struct vm_area_struct *vma);
 extern int remap_page_range(unsigned long from, unsigned long to, unsigned long size, pgprot_t prot);
 extern int zeromap_page_range(unsigned long from, unsigned long size, pgprot_t prot);
-extern int copy_page_range(struct mm_struct *dst, struct mm_struct *src, struct vm_area_struct *vma);
-extern int zap_page_range(struct mm_struct *mm, unsigned long address, unsigned long size);
 
+extern void vmtruncate(struct inode * inode, unsigned long offset);
 extern void handle_mm_fault(struct vm_area_struct *vma, unsigned long address, int write_access);
 extern void do_wp_page(struct task_struct * tsk, struct vm_area_struct * vma, unsigned long address, int write_access);
 extern void do_no_page(struct task_struct * tsk, struct vm_area_struct * vma, unsigned long address, int write_access);
@@ -215,8 +207,6 @@ extern void rw_swap_page(int rw, unsigned long nr, char * buf);
 /* mmap.c */
 extern unsigned long do_mmap(struct file * file, unsigned long addr, unsigned long len,
 	unsigned long prot, unsigned long flags, unsigned long off);
-extern struct vm_area_struct * find_vma (struct task_struct *, unsigned long);
-extern struct vm_area_struct * find_vma_intersection (struct task_struct *, unsigned long, unsigned long);
 extern void merge_segments(struct task_struct *, unsigned long, unsigned long);
 extern void insert_vm_struct(struct task_struct *, struct vm_area_struct *);
 extern void remove_shared_vm_struct(struct vm_area_struct *);
@@ -225,6 +215,10 @@ extern void exit_mmap(struct mm_struct *);
 extern int do_munmap(unsigned long, size_t);
 extern unsigned long get_unmapped_area(unsigned long, unsigned long);
 
+/* filemap.c */
+extern unsigned long page_unuse(unsigned long);
+extern int shrink_mmap(int, unsigned long);
+
 #define read_swap_page(nr,buf) \
 	rw_swap_page(READ,(nr),(buf))
 #define write_swap_page(nr,buf) \
@@ -244,6 +238,56 @@ extern unsigned long get_unmapped_area(unsigned long, unsigned long);
 
 #define GFP_LEVEL_MASK 0xf
 
+#define avl_empty	(struct vm_area_struct *) NULL
+
+static inline int expand_stack(struct vm_area_struct * vma, unsigned long address)
+{
+	unsigned long grow;
+
+	address &= PAGE_MASK;
+	if (vma->vm_end - address > current->rlim[RLIMIT_STACK].rlim_cur)
+		return -ENOMEM;
+	grow = vma->vm_start - address;
+	vma->vm_start = address;
+	vma->vm_offset -= grow;
+	vma->vm_mm->total_vm += grow >> PAGE_SHIFT;
+	if (vma->vm_flags & VM_LOCKED)
+		vma->vm_mm->locked_vm += grow >> PAGE_SHIFT;
+	return 0;
+}
+
+/* Look up the first VMA which satisfies  addr < vm_end,  NULL if none. */
+static inline struct vm_area_struct * find_vma (struct task_struct * task, unsigned long addr)
+{
+	struct vm_area_struct * result = NULL;
+	struct vm_area_struct * tree;
+
+	if (!task->mm)
+		return NULL;
+	for (tree = task->mm->mmap_avl ; ; ) {
+		if (tree == avl_empty)
+			return result;
+		if (tree->vm_end > addr) {
+			if (tree->vm_start <= addr)
+				return tree;
+			result = tree;
+			tree = tree->vm_avl_left;
+		} else
+			tree = tree->vm_avl_right;
+	}
+}
+
+/* Look up the first VMA which intersects the interval start_addr..end_addr-1,
+   NULL if none.  Assume start_addr < end_addr. */
+static inline struct vm_area_struct * find_vma_intersection (struct task_struct * task, unsigned long start_addr, unsigned long end_addr)
+{
+	struct vm_area_struct * vma;
+
+	vma = find_vma(task,start_addr);
+	if (!vma || end_addr <= vma->vm_start)
+		return NULL;
+	return vma;
+}
 
 /*
  * vm_ops not present page codes for shared memory.
