@@ -33,7 +33,6 @@
  * 		Idea by Alex Bligh (alex@cconcepts.co.uk)
  */
 
-#include <linux/config.h>
 #include <linux/signal.h>
 #include <linux/sched.h>
 #include <linux/head.h>
@@ -58,9 +57,21 @@ unsigned long high_memory = 0;
 int nr_swap_pages = 0;
 int nr_free_pages = 0;
 struct mem_list free_area_list[NR_MEM_LISTS];
-unsigned char * free_area_map[NR_MEM_LISTS];
+unsigned int * free_area_map[NR_MEM_LISTS];
 
-#define copy_page(from,to) memcpy((void *) to, (void *) from, PAGE_SIZE)
+/*
+ * We special-case the C-O-W ZERO_PAGE, because it's such
+ * a common occurrence (no need to read the page to know
+ * that it's zero - better for the cache and memory subsystem).
+ */
+static inline void copy_page(unsigned long from, unsigned long to)
+{
+	if (from == ZERO_PAGE) {
+		memset((void *) to, 0, PAGE_SIZE);
+		return;
+	}
+	memcpy((void *) to, (void *) from, PAGE_SIZE);
+}
 
 #define USER_PTRS_PER_PGD (TASK_SIZE / PGDIR_SIZE)
 
@@ -119,60 +130,30 @@ static inline void free_one_pgd(pgd_t * dir)
 	pmd_free(pmd);
 }
 	
-int new_page_tables(struct task_struct * tsk)
-{
-	pgd_t * page_dir, * new_pg;
-	int i;
-
-	if (!(new_pg = pgd_alloc()))
-		return -ENOMEM;
-	page_dir = pgd_offset(&init_mm, 0);
-	for (i = USER_PTRS_PER_PGD ; i < PTRS_PER_PGD ; i++)
-		new_pg[i] = page_dir[i];
-	SET_PAGE_DIR(tsk, new_pg);
-	tsk->mm->pgd = new_pg;
-	return 0;
-}
-
 /*
  * This function clears all user-level page tables of a process - this
- * is needed by execve(), so that old pages aren't in the way. Note that
- * unlike 'free_page_tables()', this function still leaves a valid
- * page-table-tree in memory: it just removes the user pages. The two
- * functions are similar, but there is a fundamental difference.
+ * is needed by execve(), so that old pages aren't in the way.
  */
 void clear_page_tables(struct task_struct * tsk)
 {
 	int i;
 	pgd_t * page_dir;
 
-	if (!tsk)
-		return;
-	if (tsk == task[0])
-		panic("task[0] (swapper) doesn't support exec()\n");
-	page_dir = pgd_offset(tsk->mm, 0);
-	if (!page_dir) {
-		printk("%s trying to clear NULL page-directory: not good\n", tsk->comm);
-		return;
-	}
-	if (pgd_inuse(page_dir)) {
-		if (new_page_tables(tsk))
-			oom(tsk);
-		pgd_free(page_dir);
-		return;
-	}
-	if (page_dir == swapper_pg_dir) {
+	page_dir = tsk->mm->pgd;
+	if (!page_dir || page_dir == swapper_pg_dir) {
 		printk("%s trying to clear kernel page-directory: not good\n", tsk->comm);
 		return;
 	}
 	for (i = 0 ; i < USER_PTRS_PER_PGD ; i++)
 		free_one_pgd(page_dir + i);
-	invalidate();
-	return;
+	invalidate_mm(tsk->mm);
 }
 
 /*
- * This function frees up all page tables of a process when it exits.
+ * This function frees up all page tables of a process when it exits. It
+ * is the same as "clear_page_tables()", except it also changes the process'
+ * page table directory to the kernel page tables and then frees the old
+ * page table directory.
  */
 void free_page_tables(struct task_struct * tsk)
 {
@@ -184,16 +165,28 @@ void free_page_tables(struct task_struct * tsk)
 		printk("%s trying to free kernel page-directory: not good\n", tsk->comm);
 		return;
 	}
+	invalidate_mm(tsk->mm);
 	SET_PAGE_DIR(tsk, swapper_pg_dir);
-	if (pgd_inuse(page_dir)) {
-		pgd_free(page_dir);
-		return;
-	}
 	tsk->mm->pgd = swapper_pg_dir;	/* or else... */
 	for (i = 0 ; i < PTRS_PER_PGD ; i++)
 		free_one_pgd(page_dir + i);
 	pgd_free(page_dir);
-	invalidate();
+}
+
+int new_page_tables(struct task_struct * tsk)
+{
+	pgd_t * page_dir, * new_pg;
+	int i;
+
+	if (!(new_pg = pgd_alloc()))
+		return -ENOMEM;
+	page_dir = pgd_offset(&init_mm, 0);
+	for (i = USER_PTRS_PER_PGD ; i < PTRS_PER_PGD ; i++)
+		new_pg[i] = page_dir[i];
+	invalidate_mm(tsk->mm);
+	SET_PAGE_DIR(tsk, new_pg);
+	tsk->mm->pgd = new_pg;
+	return 0;
 }
 
 static inline void copy_one_pte(pte_t * old_pte, pte_t * new_pte)
@@ -207,7 +200,7 @@ static inline void copy_one_pte(pte_t * old_pte, pte_t * new_pte)
 		set_pte(new_pte, pte);
 		return;
 	}
-	if (pte_page(pte) > high_memory || (mem_map[MAP_NR(pte_page(pte))] & MAP_PAGE_RESERVED)) {
+	if (pte_page(pte) > high_memory || mem_map[MAP_NR(pte_page(pte))].reserved) {
 		set_pte(new_pte, pte);
 		return;
 	}
@@ -217,7 +210,7 @@ static inline void copy_one_pte(pte_t * old_pte, pte_t * new_pte)
 		pte = pte_mkdirty(pte);
 	set_pte(new_pte, pte_mkold(pte));
 	set_pte(old_pte, pte);
-	mem_map[MAP_NR(pte_page(pte))]++;
+	mem_map[MAP_NR(pte_page(pte))].count++;
 }
 
 static inline int copy_pte_range(pmd_t *dst_pmd, pmd_t *src_pmd, unsigned long address, unsigned long size)
@@ -305,7 +298,9 @@ int copy_page_range(struct mm_struct *dst, struct mm_struct *src,
 			break;
 		address = (address + PGDIR_SIZE) & PGDIR_MASK;
 	}
-	invalidate();
+	/* Note that the src ptes get c-o-w treatment, so they change too. */
+	invalidate_range(src, vma->vm_start, vma->vm_end);
+	invalidate_range(dst, vma->vm_start, vma->vm_end);
 	return error;
 }
 
@@ -314,9 +309,10 @@ static inline void forget_pte(pte_t page)
 	if (pte_none(page))
 		return;
 	if (pte_present(page)) {
-		free_page(pte_page(page));
-		if (mem_map[MAP_NR(pte_page(page))] & MAP_PAGE_RESERVED)
+		unsigned long addr = pte_page(page);
+		if (addr >= high_memory || mem_map[MAP_NR(addr)].reserved)
 			return;
+		free_page(addr);
 		if (current->mm->rss <= 0)
 			return;
 		current->mm->rss--;
@@ -325,7 +321,7 @@ static inline void forget_pte(pte_t page)
 	swap_free(pte_val(page));
 }
 
-static inline void unmap_pte_range(pmd_t * pmd, unsigned long address, unsigned long size)
+static inline void zap_pte_range(pmd_t * pmd, unsigned long address, unsigned long size)
 {
 	pte_t * pte;
 	unsigned long end;
@@ -333,7 +329,7 @@ static inline void unmap_pte_range(pmd_t * pmd, unsigned long address, unsigned
 	if (pmd_none(*pmd))
 		return;
 	if (pmd_bad(*pmd)) {
-		printk("unmap_pte_range: bad pmd (%08lx)\n", pmd_val(*pmd));
+		printk("zap_pte_range: bad pmd (%08lx)\n", pmd_val(*pmd));
 		pmd_clear(pmd);
 		return;
 	}
@@ -351,7 +347,7 @@ static inline void unmap_pte_range(pmd_t * pmd, unsigned long address, unsigned
 	} while (address < end);
 }
 
-static inline void unmap_pmd_range(pgd_t * dir, unsigned long address, unsigned long size)
+static inline void zap_pmd_range(pgd_t * dir, unsigned long address, unsigned long size)
 {
 	pmd_t * pmd;
 	unsigned long end;
@@ -359,7 +355,7 @@ static inline void unmap_pmd_range(pgd_t * dir, unsigned long address, unsigned
 	if (pgd_none(*dir))
 		return;
 	if (pgd_bad(*dir)) {
-		printk("unmap_pmd_range: bad pgd (%08lx)\n", pgd_val(*dir));
+		printk("zap_pmd_range: bad pgd (%08lx)\n", pgd_val(*dir));
 		pgd_clear(dir);
 		return;
 	}
@@ -369,7 +365,7 @@ static inline void unmap_pmd_range(pgd_t * dir, unsigned long address, unsigned
 	if (end > PGDIR_SIZE)
 		end = PGDIR_SIZE;
 	do {
-		unmap_pte_range(pmd, address, end - address);
+		zap_pte_range(pmd, address, end - address);
 		address = (address + PMD_SIZE) & PMD_MASK; 
 		pmd++;
 	} while (address < end);
@@ -385,23 +381,14 @@ int zap_page_range(struct mm_struct *mm, unsigned long address, unsigned long si
 
 	dir = pgd_offset(mm, address);
 	while (address < end) {
-		unmap_pmd_range(dir, address, end - address);
+		zap_pmd_range(dir, address, end - address);
 		address = (address + PGDIR_SIZE) & PGDIR_MASK;
 		dir++;
 	}
-	invalidate();
+	invalidate_range(mm, end - size, end);
 	return 0;
 }
 
-/*
- * a more complete version of free_page_tables which performs with page
- * granularity.
- */
-int unmap_page_range(unsigned long address, unsigned long size)
-{
-	return zap_page_range(current->mm, address, size);
-}
-
 static inline void zeromap_pte_range(pte_t * pte, unsigned long address, unsigned long size, pte_t zero_pte)
 {
 	unsigned long end;
@@ -458,7 +445,7 @@ int zeromap_page_range(unsigned long address, unsigned long size, pgprot_t prot)
 		address = (address + PGDIR_SIZE) & PGDIR_MASK;
 		dir++;
 	}
-	invalidate();
+	invalidate_range(current->mm, end - size, end);
 	return error;
 }
 
@@ -479,7 +466,7 @@ static inline void remap_pte_range(pte_t * pte, unsigned long address, unsigned
 	do {
 		pte_t oldpage = *pte;
 		pte_clear(pte);
-		if (offset >= high_memory || (mem_map[MAP_NR(offset)] & MAP_PAGE_RESERVED))
+		if (offset >= high_memory || mem_map[MAP_NR(offset)].reserved)
 			set_pte(pte, mk_pte(offset, prot));
 		forget_pte(oldpage);
 		address += PAGE_SIZE;
@@ -528,7 +515,7 @@ int remap_page_range(unsigned long from, unsigned long offset, unsigned long siz
 		from = (from + PGDIR_SIZE) & PGDIR_MASK;
 		dir++;
 	}
-	invalidate();
+	invalidate_range(current->mm, from - size, from);
 	return error;
 }
 
@@ -543,7 +530,7 @@ static void put_page(pte_t * page_table, pte_t pte)
 		return;
 	}
 /* no need for invalidate */
-	*page_table = pte;
+	set_pte(page_table, pte);
 }
 
 /*
@@ -558,7 +545,7 @@ unsigned long put_dirty_page(struct task_struct * tsk, unsigned long page, unsig
 
 	if (page >= high_memory)
 		printk("put_dirty_page: trying to put page %08lx at %08lx\n",page,address);
-	if (mem_map[MAP_NR(page)] != 1)
+	if (mem_map[MAP_NR(page)].count != 1)
 		printk("mem_map disagrees with %08lx at %08lx\n",page,address);
 	pgd = pgd_offset(tsk->mm,address);
 	pmd = pmd_alloc(pgd, address);
@@ -575,8 +562,8 @@ unsigned long put_dirty_page(struct task_struct * tsk, unsigned long page, unsig
 	}
 	if (!pte_none(*pte)) {
 		printk("put_dirty_page: page already exists\n");
-		pte_clear(pte);
-		invalidate();
+		free_page(page);
+		return 0;
 	}
 	set_pte(pte, pte_mkwrite(pte_mkdirty(mk_pte(page, PAGE_COPY))));
 /* no need for invalidate */
@@ -628,28 +615,28 @@ void do_wp_page(struct task_struct * tsk, struct vm_area_struct * vma,
 	old_page = pte_page(pte);
 	if (old_page >= high_memory)
 		goto bad_wp_page;
-	vma->vm_mm->min_flt++;
+	tsk->min_flt++;
 	/*
 	 * Do we need to copy?
 	 */
-	if (mem_map[MAP_NR(old_page)] != 1) {
+	if (mem_map[MAP_NR(old_page)].count != 1) {
 		if (new_page) {
-			if (mem_map[MAP_NR(old_page)] & MAP_PAGE_RESERVED)
+			if (mem_map[MAP_NR(old_page)].reserved)
 				++vma->vm_mm->rss;
 			copy_page(old_page,new_page);
 			set_pte(page_table, pte_mkwrite(pte_mkdirty(mk_pte(new_page, vma->vm_page_prot))));
 			free_page(old_page);
-			invalidate();
+			invalidate_page(vma, address);
 			return;
 		}
 		set_pte(page_table, BAD_PAGE);
 		free_page(old_page);
 		oom(tsk);
-		invalidate();
+		invalidate_page(vma, address);
 		return;
 	}
 	set_pte(page_table, pte_mkdirty(pte_mkwrite(pte)));
-	invalidate();
+	invalidate_page(vma, address);
 	if (new_page)
 		free_page(new_page);
 	return;
@@ -682,7 +669,7 @@ int verify_area(int type, const void * addr, unsigned long size)
 	 * case where we use a fake user buffer with get_fs/set_fs()) we
 	 * don't expect to find the address in the user vm map.
 	 */
-	if (get_fs() == get_ds())
+	if (!size || get_fs() == get_ds())
 		return 0;
 
 	vma = find_vma(current, start);
@@ -692,7 +679,7 @@ int verify_area(int type, const void * addr, unsigned long size)
 		goto good_area;
 	if (!(vma->vm_flags & VM_GROWSDOWN))
 		goto bad_area;
-	if (vma->vm_end - start > current->rlim[RLIMIT_STACK].rlim_cur)
+	if (expand_stack(vma, start))
 		goto bad_area;
 
 good_area:
@@ -812,7 +799,7 @@ static int try_to_share(unsigned long to_address, struct vm_area_struct * to_are
 /* is the page reasonable at all? */
 	if (pte_page(from) >= high_memory)
 		return 0;
-	if (mem_map[MAP_NR(pte_page(from))] & MAP_PAGE_RESERVED)
+	if (mem_map[MAP_NR(pte_page(from))].reserved)
 		return 0;
 /* is the destination ok? */
 	to_dir = pgd_offset(to_area->vm_mm,to_address);
@@ -862,7 +849,7 @@ static int try_to_share(unsigned long to_address, struct vm_area_struct * to_are
 		set_pte(from_table, pte_mkdirty(from));
 		delete_from_swap_cache(pte_page(from));
 	}
-	mem_map[MAP_NR(pte_page(from))]++;
+	mem_map[MAP_NR(pte_page(from))].count++;
 	set_pte(to_table, mk_pte(pte_page(from), to_area->vm_page_prot));
 /* Check if we need to do anything at all to the 'from' field */
 	if (!pte_write(from))
@@ -871,7 +858,7 @@ static int try_to_share(unsigned long to_address, struct vm_area_struct * to_are
 		return 1;
 /* ok, need to mark it read-only, so invalidate any possible old TB entry */
 	set_pte(from_table, pte_wrprotect(from));
-	invalidate();
+	invalidate_page(from_area, from_address);
 	return 1;
 }
 
@@ -928,6 +915,83 @@ static int share_page(struct vm_area_struct * area, unsigned long address,
 	return 0;
 }
 
+/*
+ * This function zeroes out partial mmap'ed pages at truncation time..
+ */
+static void partial_clear(struct vm_area_struct *vma, unsigned long address)
+{
+	pgd_t *page_dir;
+	pmd_t *page_middle;
+	pte_t *page_table, pte;
+
+	page_dir = pgd_offset(vma->vm_mm, address);
+	if (pgd_none(*page_dir))
+		return;
+	if (pgd_bad(*page_dir)) {
+		printk("bad page table directory entry %p:[%lx]\n", page_dir, pgd_val(*page_dir));
+		pgd_clear(page_dir);
+		return;
+	}
+	page_middle = pmd_offset(page_dir, address);
+	if (pmd_none(*page_middle))
+		return;
+	if (pmd_bad(*page_middle)) {
+		printk("bad page table directory entry %p:[%lx]\n", page_dir, pgd_val(*page_dir));
+		pmd_clear(page_middle);
+		return;
+	}
+	page_table = pte_offset(page_middle, address);
+	pte = *page_table;
+	if (!pte_present(pte))
+		return;
+	address &= ~PAGE_MASK;
+	address += pte_page(pte);
+	if (address >= high_memory)
+		return;
+	memset((void *) address, 0, PAGE_SIZE - (address & ~PAGE_MASK));
+}
+
+/*
+ * Handle all mappings that got truncated by a "truncate()"
+ * system call.
+ *
+ * NOTE! We have to be ready to update the memory sharing
+ * between the file and the memory map for a potential last
+ * incomplete page.  Ugly, but necessary.
+ */
+void vmtruncate(struct inode * inode, unsigned long offset)
+{
+	struct vm_area_struct * mpnt;
+
+	invalidate_inode_pages(inode, offset);
+	if (!inode->i_mmap)
+		return;
+	mpnt = inode->i_mmap;
+	do {
+		unsigned long start = mpnt->vm_start;
+		unsigned long len = mpnt->vm_end - start;
+		unsigned long diff;
+
+		/* mapping wholly truncated? */
+		if (mpnt->vm_offset >= offset) {
+			zap_page_range(mpnt->vm_mm, start, len);
+			continue;
+		}
+		/* mapping wholly unaffected? */
+		diff = offset - mpnt->vm_offset;
+		if (diff >= len)
+			continue;
+		/* Ok, partially affected.. */
+		start += diff;
+		len = (len - diff) & PAGE_MASK;
+		if (start & ~PAGE_MASK) {
+			partial_clear(mpnt, start);
+			start = (start + ~PAGE_MASK) & PAGE_MASK;
+		}
+		zap_page_range(mpnt->vm_mm, start, len);
+	} while ((mpnt = mpnt->vm_next_share) != inode->i_mmap);
+}
+
 /*
  * fill in an empty page-table if none exists.
  */
@@ -966,10 +1030,10 @@ static inline void do_swap_page(struct task_struct * tsk,
 		free_page(pte_page(page));
 		return;
 	}
-	if (mem_map[MAP_NR(pte_page(page))] > 1 && !(vma->vm_flags & VM_SHARED))
+	if (mem_map[MAP_NR(pte_page(page))].count > 1 && !(vma->vm_flags & VM_SHARED))
 		page = pte_wrprotect(page);
 	++vma->vm_mm->rss;
-	++vma->vm_mm->maj_flt;
+	++tsk->maj_flt;
 	set_pte(page_table, page);
 	return;
 }
@@ -1000,14 +1064,14 @@ void do_no_page(struct task_struct * tsk, struct vm_area_struct * vma,
 	address &= PAGE_MASK;
 	if (!vma->vm_ops || !vma->vm_ops->nopage) {
 		++vma->vm_mm->rss;
-		++vma->vm_mm->min_flt;
+		++tsk->min_flt;
 		get_empty_page(tsk, vma, page_table);
 		return;
 	}
 	page = __get_free_page(GFP_KERNEL);
 	if (share_page(vma, address, write_access, page)) {
-		++vma->vm_mm->min_flt;
 		++vma->vm_mm->rss;
+		++tsk->min_flt;
 		return;
 	}
 	if (!page) {
@@ -1015,7 +1079,7 @@ void do_no_page(struct task_struct * tsk, struct vm_area_struct * vma,
 		put_page(page_table, BAD_PAGE);
 		return;
 	}
-	++vma->vm_mm->maj_flt;
+	++tsk->maj_flt;
 	++vma->vm_mm->rss;
 	/*
 	 * The fourth argument is "no_share", which tells the low-level code
@@ -1041,7 +1105,7 @@ void do_no_page(struct task_struct * tsk, struct vm_area_struct * vma,
 	entry = mk_pte(page, vma->vm_page_prot);
 	if (write_access) {
 		entry = pte_mkwrite(pte_mkdirty(entry));
-	} else if (mem_map[MAP_NR(page)] > 1 && !(vma->vm_flags & VM_SHARED))
+	} else if (mem_map[MAP_NR(page)].count > 1 && !(vma->vm_flags & VM_SHARED))
 		entry = pte_wrprotect(entry);
 	put_page(page_table, entry);
 }
